# -*- coding: utf-8 -*-

"""
Collection of helper functions being used in lbsntransform package.
"""


import csv
import datetime as dt
import logging
import re
import sys
import time
from datetime import timezone
from json import JSONDecodeError, JSONDecoder

import emoji
from google.protobuf.timestamp_pb2 import Timestamp
from lbsnstructure.lbsnstructure_pb2 import (CompositeKey, RelationshipKey,
                                             lbsnCity, lbsnCountry, lbsnPlace,
                                             lbsnPost, lbsnPostReaction,
                                             lbsnRelationship, lbsnUser,
                                             lbsnUserGroup)
from shapely import geos, wkt

# pylint: disable=no-member


class HelperFunctions():

    @staticmethod
    def set_logger():
        """ Set logging handler manually,
        so we can also print to console while logging to file
        """

        logging.basicConfig(
            handlers=[logging.FileHandler(
                'log.log', 'w', 'utf-8')],
            format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
            datefmt='%H:%M:%S',
            level=logging.DEBUG)
        log = logging.getLogger(__name__)

        # Get Stream handler
        logging.getLogger().addHandler(logging.StreamHandler())
        return log

    @staticmethod
    def geoacc_within_threshold(post_geoaccuracy, min_geoaccuracy):
        """Checks if geoaccuracy is within or below threshhold defined"""
        if min_geoaccuracy == lbsnPost.LATLNG:
            allowed_geoaccuracies = [lbsnPost.LATLNG]
        elif min_geoaccuracy == lbsnPost.PLACE:
            allowed_geoaccuracies = [lbsnPost.LATLNG, lbsnPost.PLACE]
        elif min_geoaccuracy == lbsnPost.CITY:
            allowed_geoaccuracies = [lbsnPost.LATLNG, lbsnPost.PLACE,
                                     lbsnPost.CITY]
        else:
            return True
        # check post geoaccuracy
        if post_geoaccuracy in allowed_geoaccuracies:
            return True
        else:
            return False

    @staticmethod
    def get_version():
        """Gets the program version number from version file in root"""
        with open('VERSION') as version_file:
            version_var = version_file.read().strip()
        return version_var

    @staticmethod
    def log_main_debug(debug_text):
        """Issues a main debug log in case it is
        needed for static functions.
        """
        logging.getLogger('__main__').debug(debug_text)

    @staticmethod
    def null_notice(x):
        """Reporting: Suppresses null notice (for Null island)
        if value is zero.
        """
        def null_notice_x(x): return f'(Null Island: {x})' if x > 0 else ''
        return null_notice_x(x)

    @staticmethod
    def utc_to_local(utc_dt):
        return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)

    @staticmethod
    def cleanhtml(raw_html):
        cleanr = re.compile('<.*?>')
        cleantext = re.sub(cleanr, '', raw_html)
        return cleantext

    @staticmethod
    def extract_emoji(str):
        # str = str.decode('utf-32').encode('utf-32', 'surrogatepass')
        # return list(c for c in str if c in emoji.UNICODE_EMOJI)
        return list(c for c in str if c in emoji.UNICODE_EMOJI)

    @staticmethod
    def get_rectangle_bounds(points):
        lats = []
        lngs = []
        for point in points:
            lngs.append(point[0])
            lats.append(point[1])
        lim_y_min = min(lats)
        lim_y_max = max(lats)
        lim_x_min = min(lngs)
        lim_x_max = max(lngs)
        return lim_y_min, lim_y_max, lim_x_min, lim_x_max

    @staticmethod
    def new_lbsn_record_with_id(record, id, origin):
            # initializes new record with composite ID
        c_key = CompositeKey()
        c_key.origin.CopyFrom(origin)
        c_key.id = id
        record.pkey.CopyFrom(c_key)
        return record

    @staticmethod
    def new_lbsn_relation_with_id(lbsn_relationship,
                                  relation_to_id,
                                  relation_from_id,
                                  relation_origin):
        # initializes new relationship with 2 composite IDs for one origin
        c_key_to = CompositeKey()
        c_key_to.origin.CopyFrom(relation_origin)
        c_key_to.id = relation_to_id
        c_key_from = CompositeKey()
        c_key_from.origin.CopyFrom(relation_origin)
        c_key_from.id = relation_from_id
        r_key = RelationshipKey()
        r_key.relation_to.CopyFrom(c_key_to)
        r_key.relation_from.CopyFrom(c_key_from)
        lbsn_relationship.pkey.CopyFrom(r_key)
        return lbsn_relationship

    @staticmethod
    def is_post_reaction(jsonString):
        if 'quoted_status' in jsonString \
                or 'retweeted_status' in jsonString \
                or jsonString.get('in_reply_to_status_id_str'):
            # The retweeted field will return true if a tweet _got_ retweeted
            # To detect if a tweet is a retweet of other tweet,
            # check the retweeted_status field
            return True
        else:
            return False

    @staticmethod
    def assign_media_post_type(json_media_string):
        # if post, get type of first entity
        type_string = json_media_string[0].get('type')
        # type is either photo, video, or animated_gif
        # https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/extended-entities-object.html
        if type_string:
            if type_string == "photo":
                post_type = lbsnPost.IMAGE
            elif type_string in ("video", "animated_gif"):
                post_type = lbsnPost.VIDEO

        else:
            post_type = lbsnPost.OTHER
            logging.getLogger('__main__').debug(f'Other Post type detected: '
                                                f'{json_media_string}')
        return post_type

    @staticmethod
    def json_date_string_to_proto(json_date_string):
        # Parse String -Timestamp Format found in Twitter json
        date_time_record = dt.datetime.strptime(json_date_string,
                                                '%a %b %d %H:%M:%S +0000 %Y')
        protobuf_timestamp_record = Timestamp()
        # Convert to ProtoBuf Timestamp Recommendation
        protobuf_timestamp_record.FromDatetime(date_time_record)
        return protobuf_timestamp_record

    @staticmethod
    def json_date_timestamp_to_proto(json_date_timestamp):
        # Parse String -Timestamp Format found in Twitter json
        date_time_record = dt.datetime.fromtimestamp(json_date_timestamp)
        protobuf_timestamp_record = Timestamp()
        # Convert to ProtoBuf Timestamp Recommendation
        protobuf_timestamp_record.FromDatetime(date_time_record)
        return protobuf_timestamp_record

    @staticmethod
    def parse_csv_datestring_to_protobuf(csv_datestring, t_format=None):
        """Parse String -Timestamp Format found in Flickr csv

        e.g. 2012-02-16 09:56:37.0
        """
        if t_format is None:
            t_format = '%m/%d/%Y %H:%M:%S'
        try:
            date_time_record = dt.datetime.strptime(
                csv_datestring, t_format)
        except ValueError:
            return None
        protobuf_timestamp_record = Timestamp()
        # Convert to ProtoBuf Timestamp Recommendation
        protobuf_timestamp_record.FromDatetime(date_time_record)
        return protobuf_timestamp_record

    @staticmethod
    def parse_timestamp_string_to_protobuf(timestamp_string):
        """Converts timestamp string to protobuf object"""
        # Parse from RFC 3339 date string to Timestamp.
        time_date = dt.datetime.fromtimestamp(int(timestamp_string))
        protobuf_timestamp_record = Timestamp()
        protobuf_timestamp_record.FromDatetime(time_date)
        return protobuf_timestamp_record

    @staticmethod
    def get_mentioned_users(userMentions_jsonString, origin):
        mentioned_users_list = []
        for user_mention in userMentions_jsonString:  # iterate over the list
            ref_user_record = \
                HelperFunctions.new_lbsn_record_with_id(lbsnUser(),
                                                        user_mention.get(
                                                        'id_str'),
                                                        origin)
            ref_user_record.user_fullname = \
                user_mention.get('name')  # Needs to be saved
            ref_user_record.user_name = user_mention.get('screen_name')
            mentioned_users_list.append(ref_user_record)
        return mentioned_users_list

    @staticmethod
    def substitute_referenced_user(main_post, origin, log):
        # Look for mentioned userRecords
        ref_user_pkey = None
        user_mentions_json = main_post.get('entities').get('user_mentions')
        if user_mentions_json:
            ref_user_records = \
                HelperFunctions.get_mentioned_users(user_mentions_json,
                                                    origin)
            # if it is a retweet, and the status contains 'RT @',
            # and the mentioned UserID is also in the status,
            # we can almost be completely certain that it is the userid who
            # posted the original tweet that was retweeted
            if ref_user_records \
               and ref_user_records[0].user_name.lower() in \
               main_post.get('text').lower() \
               and main_post.get('text').startswith(f'RT @'):
                ref_user_pkey = ref_user_records[0].pkey
            if ref_user_pkey is None:
                log.warning(
                    f'No User record found for referenced post in: '
                    f'{main_post}')
                input("Press Enter to continue... "
                      "(post will be saved without userid)")
        return ref_user_pkey

    @staticmethod
    def null_check(record_attr):
        """Helper function to check for Null Values
        """
        if not record_attr:
            # will catch empty and None
            return None
        # This function will also remove Null bytes from string,
        # which aren't supported by Postgres
        if isinstance(record_attr, str):
            record_attr = HelperFunctions.clean_null_bytes_from_str(
                record_attr)
        return record_attr

    @staticmethod
    def null_geom_check(geom_attr):
        """Helper function to check for Null Values
        in geometry columns and replace with Null Island

        Note:
        null_geom_check is only applied to geometry columns
        with NOT NULL Constraint
        """
        if geom_attr is None or \
                (isinstance(geom_attr, str) and geom_attr == ''):
            null_island = "POINT(%s %s)" % (0, 0)
            return null_island
        return geom_attr

    @staticmethod
    def null_check_datetime(recordAttr):
        if not recordAttr:
            # will catch empty and None
            return None
        try:
            dt_attr = recordAttr.ToDatetime()
        except:
            return None
        if dt_attr == dt.datetime(1970, 1, 1, 0, 0, 0):
            return None
        else:
            return recordAttr.ToDatetime()

    @staticmethod
    def return_ewkb_from_geotext(text):
        """Generates Geometry in Well-known-Text format
        from PostGis Format (e.g. 'POINT(0 0)')
        with SRID for WGS1984 (4326)

        Note that:
        geos.WKBWriter.defaults['include_srid'] = True
        must be set (see config.py)
        """
        if text is None:
            # keep Null geometries, e.g. for geom_area columns
            return None
        geom = wkt.loads(text)
        # Set SRID to WGS1984
        geos.lgeos.GEOSSetSRID(geom._geom, 4326)
        geom = geom.wkb_hex
        return geom

    @staticmethod
    def decode_stacked(document, pos=0, decoder=JSONDecoder()):
        NOT_WHITESPACE = re.compile(r'[^\s]')
        while True:
            match = NOT_WHITESPACE.search(document, pos)
            if not match:
                return
            pos = match.start()

            try:
                obj, pos = decoder.raw_decode(document, pos)
            except JSONDecodeError:
                # do something sensible if there's some error
                raise
            yield obj

    @staticmethod
    def clean_null_bytes_from_str(str):
        str_without_null_byte = str.replace('\x00', '')
        return str_without_null_byte

    @staticmethod
    def turn_lower(str):
        """Returns lower but keeps none values"""
        if str:
            return (str.lower())
        else:
            return str

    @staticmethod
    def empty_list(list):
        """Returns lower but keeps none values"""
        if list:
            if len(list) == 0:
                return None
            else:
                return list
        else:
            return None

    @staticmethod
    def merge_existing_records(oldrecord, newrecord):
        # Basic Compare function for GUIDS
        # First check if length of both ProtoBuf Messages are the same
        old_record_string = oldrecord.SerializeToString()
        new_record_string = newrecord.SerializeToString()
        if not len(old_record_string) == len(new_record_string):
            # no need to do anything if same lengt
            oldrecord.MergeFrom(newrecord)
            # updatedrecord = self.deepCompareMergeMessages(oldrecord,
            #                                               newrecord)

    @staticmethod
    def load_importer_mapping_module(origin):
        """ Switch import module based on origin input
            1 - Instagram, 2 - Flickr, 3 - Twitter
        """
        if origin == 2:
            from .field_mapping_flickr import FieldMappingFlickr as importer
        if origin == 21:
            # Flickr YFCC100M dataset
            from .field_mapping_yfcc100m import FieldMappingYFCC100M as importer
        elif origin == 3:
            from .field_mapping_twitter import FieldMappingTwitter as importer
        return importer

    @staticmethod
    def dict_type_switcher(desc_name):
        """ Create protoBuf messages by name
        """
        dict_switcher = {
            lbsnCountry().DESCRIPTOR.name: lbsnCountry(),
            lbsnCity().DESCRIPTOR.name: lbsnCity(),
            lbsnPlace().DESCRIPTOR.name: lbsnPlace(),
            lbsnUser().DESCRIPTOR.name: lbsnUser(),
            lbsnUserGroup().DESCRIPTOR.name:  lbsnUserGroup(),
            lbsnPost().DESCRIPTOR.name: lbsnPost(),
            lbsnPostReaction().DESCRIPTOR.name: lbsnPostReaction(),
            lbsnRelationship().DESCRIPTOR.name: lbsnRelationship()
        }
        return dict_switcher.get(desc_name)

    @staticmethod
    def check_notice_empty_post_guid(post_guid):
        if not post_guid:
            logging.getLogger('__main__').warning(f'No PostGuid\n\n'
                                                  f'{post_guid}')
            input("Press Enter to continue... (entry will be skipped)")
            return False
        else:
            return True


class LBSNRecordDicts():
    def __init__(self):
        self.lbsn_country_dict = dict()
        self.lbsn_city_dict = dict()
        self.lbsn_place_dict = dict()
        self.lbsn_user_group_dict = dict()
        self.lbsn_user_dict = dict()
        self.lbsn_post_dict = dict()
        self.lbsn_post_reaction_dict = dict()
        self.lbsn_relationship_dict = dict()
        self.key_hashes = {lbsnPost.DESCRIPTOR.name: set(),
                           lbsnCountry.DESCRIPTOR.name: set(),
                           lbsnCity.DESCRIPTOR.name: set(),
                           lbsnPlace.DESCRIPTOR.name: set(),
                           lbsnUserGroup.DESCRIPTOR.name: set(),
                           lbsnUser.DESCRIPTOR.name: set(),
                           lbsnPostReaction.DESCRIPTOR.name: set(),
                           lbsnRelationship.DESCRIPTOR.name: set()}
        self.count_glob = 0
        # returns all recordsDicts in correct order,
        # with names as references (tuple)
        self.all_dicts = [
            (self.lbsn_country_dict, lbsnCountry().DESCRIPTOR.name),
            (self.lbsn_city_dict, lbsnCity().DESCRIPTOR.name),
            (self.lbsn_place_dict, lbsnPlace().DESCRIPTOR.name),
            (self.lbsn_user_group_dict, lbsnUserGroup().DESCRIPTOR.name),
            (self.lbsn_user_dict, lbsnUser().DESCRIPTOR.name),
            (self.lbsn_post_dict, lbsnPost().DESCRIPTOR.name),
            (self.lbsn_post_reaction_dict, lbsnPostReaction().DESCRIPTOR.name),
            (self.lbsn_relationship_dict, lbsnRelationship().DESCRIPTOR.name)
        ]

    def get_type_counts(self):
        count_list = []
        for x, y in self.key_hashes.items():
            count_list.append(f'{x}: {len(y)} ')
        return ''.join(count_list)

    def update_key_hash(self, record):
        # Keep lists of pkeys for each type
        # this can be used to check for duplicates or to get a
        # total count for each type of records (Number of unique
        # Users, Countries, Places etc.)
        # in this case we assume that origin_id remains the same
        # in each program iteration!
        if record.DESCRIPTOR.name == lbsnRelationship().DESCRIPTOR.name:
            # we need the complete uuid of both entities for
            # relationships because they can span different origin_ids
            self.key_hashes[record.DESCRIPTOR.name].add(
                f'{record.pkey.relation_to.origin.origin_id}'
                f'{record.pkey.relation_to.id}'
                f'{record.pkey.relation_from.origin.origin_id}'
                f'{record.pkey.relation_from.id}'
                f'{record.relationship_type}')
        else:
            # all other entities can be globally uniquely
            # identified by their local guid
            self.key_hashes[record.DESCRIPTOR.name].add(record.pkey.id)

    def clear(self):
        """Clears all records from all dicts
        """
        for lbsn_dict, __ in self.all_dicts:
            lbsn_dict.clear()
        self.count_glob = 0

    def deep_compare_merge_messages(self, old_record, new_record):
        # Do a deep compare
        # ProtoBuf MergeFrom does a fine job
        # only problem is it concatenates repeate strings,
        # which may result in duplicate entries
        # we take care of this prior to submission (see submitData classes)
        old_record.MergeFrom(new_record)
        # for descriptor in oldRecord.DESCRIPTOR.fields:
        #        if descriptor.label == descriptor.LABEL_REPEATED:
        #            if value_old == value_new:
        #                return oldRecord
        #            elif not value_old:
        #                newEntries = value_new
        #            else:
        #                # only add difference (e.g. = new values)
        #                newEntries = list(set(value_new) - set(value_old))
        #            x = getattr(oldRecord, descriptor.name)
        #            x.extend(newEntries)
        return old_record

    def add_records_to_dict(self, records):
        if not records:
            return
        elif isinstance(records, (list,)):
            for record in records:
                self.add_record_to_dict(record)
        else:
            record = records
            self.add_record_to_dict(record)

    def dict_selector(self, record):
        """ Get dictionary by type name
        """
        dict_switcher = {
            lbsnPost().DESCRIPTOR.name: self.lbsn_post_dict,
            lbsnCountry().DESCRIPTOR.name: self.lbsn_country_dict,
            lbsnCity().DESCRIPTOR.name: self.lbsn_city_dict,
            lbsnPlace().DESCRIPTOR.name: self.lbsn_place_dict,
            lbsnPostReaction().DESCRIPTOR.name: self.lbsn_post_reaction_dict,
            lbsnUser().DESCRIPTOR.name: self.lbsn_user_dict,
            lbsnUserGroup().DESCRIPTOR.name: self.lbsn_user_group_dict
        }
        return dict_switcher.get(record.DESCRIPTOR.name)

    def add_record_to_dict(self, newrecord):
        sel_dict = self.dict_selector(newrecord)
        pkeyID = newrecord.pkey.id
        if newrecord.pkey.id in sel_dict:
            oldrecord = sel_dict[pkeyID]
            # oldrecord will be modified/updated
            HelperFunctions.merge_existing_records(oldrecord, newrecord)
        else:
            # just count new entries
            self.count_progress_report()
            # update keyHash only necessary for new record
            self.update_key_hash(newrecord)
            sel_dict[pkeyID] = newrecord

    def count_progress_report(self):
        self.count_glob += 1
        if self.count_glob % 1000 == 0:
            # progress report (modulo)
            print(
                f'Identified Output Records {self.count_glob}..'
                f'                                                    ',
                end='\r')
            sys.stdout.flush()

    def add_relationship_to_dict(self, newrelationship):
        pkey_id = f'{newrelationship.pkey.relation_to.origin.origin_id}' \
                  f'{newrelationship.pkey.relation_to.id}' \
                  f'{newrelationship.pkey.relation_from.origin.origin_id}' \
                  f'{newrelationship.pkey.relation_from.id}' \
                  f'{newrelationship.relationship_type}'
        if pkey_id not in self.lbsn_relationship_dict:
            self.count_progress_report()
            self.lbsn_relationship_dict[pkey_id] = newrelationship
            # update keyHash only necessary for new record
            self.update_key_hash(newrelationship)


class GeocodeLocations():
    """Class for geocoding of text to lat/lng values.
    """

    def __init__(self):
        self.geocode_dict = dict()

    def load_geocodelist(self, file):
        # read each unsorted file and sort lines based on datetime (as string)
        with open(file, newline='', encoding='utf8') as f:
            # next(f) #Skip Headerrow
            locationfile_list = csv.reader(
                f, delimiter=',', quotechar='', quoting=csv.QUOTE_NONE)
            for location_geocode in locationfile_list:
                self.geocode_dict[location_geocode[2].replace(
                    ';', ',')] = (float(location_geocode[0]),  # lat
                                  location_geocode[1])  # lng
        print(f'Loaded {len(self.geocode_dict)} geocodes.')


class TimeMonitor():
    def __init__(self):
        self.now = time.time()

    def stop_time(self):
        later = time.time()
        hours, rem = divmod(later-self.now, 3600)
        minutes, seconds = divmod(rem, 60)
        # difference = int(later - self.now)
        report_msg = f'{int(hours):0>2} Hours {int(minutes):0>2} ' \
                     f'Minutes and {seconds:05.2f} Seconds passed.'
        return report_msg


class MemoryLeakDetec():
    """Identifies memory leaks

    execute .before() and .after() to see the difference
    in new objects being added
    execute report to list
    if there are high numbers of specific types of objects,
    use printType to print these, e.g. .printType(list)
    see also http://tech.labs.oliverwyman.com/blog/
    2008/11/14/tracing-python-memory-leaks/
    """
    os = __import__('os')

    def __init__(self):
        global defaultdict  # pylint: disable=global-variable-not-assigned
        global get_objects  # pylint: disable=global-variable-not-assigned
        from collections import defaultdict
        from gc import get_objects
        self._before = defaultdict(int)
        self._after = defaultdict(int)

    def before(self):
        global get_objects  # pylint: disable=global-variable-not-assigned
        for i in get_objects():  # pylint: disable=undefined-variable
            self._before[type(i)] += 1

    def after(self):
        global get_objects  # pylint: disable=global-variable-not-assigned
        for i in get_objects():  # pylint: disable=undefined-variable
            self._after[type(i)] += 1

    def report(self):
        reportStat = ""
        if self._before and self._after:
            reportStat = [(k, self._after[k] - self._before[k])
                          for k in self._after if
                          self._after[k] - self._before[k]]
        return reportStat

    def printType(self, type, max=100):
        x = 0
        toplist = []
        for obj in get_objects():  # pylint: disable=undefined-variable
            if isinstance(obj, type):
                x += 1
                if x < max:
                    toplist.append(obj)
        print(f'Count all of Type {type}: {x}. Top {max}: {toplist}')



"""
LBSNTransform: Convert Raw Social Media data
         to common LBSN interchange format (ProtoBuf) and
         transfer to local CSV or LBSN Postgres Database
"""

from __future__ import absolute_import

__author__ = "Alexander Dunkel"
__license__ = "GNU GPLv3"

import io
import logging
import sys

from .classes.helper_functions import HelperFunctions as HF
from .classes.helper_functions import LBSNRecordDicts
from .classes.load_data import LoadData
from .classes.submit_data import LBSNTransfer
from .config.config import BaseConfig


class LBSNTransform():
    """Import, convert and export RAW Location Based Social Media data,
    such as Twitter and Flickr, based on a common data structure concept
    in Google's ProtoBuf format (see package lbsnstructure).

    Input can be:
        - local CSV or Json (stacked/regular)
        - Postgres DB connection
    Output can be:
        - local CSV
        - local file with ProtoBuf encoded records
        - local SQL file ready for "Impoort from" in Postgres LBSN db
        - Postgres DB connection

    Parameters
    ----------

    origin : str, optional (default=3)
        Type of input source. Each input source has its own import mapper
        defined in a class. Feel free to add or modify classes based
        on your needs. Pre-provided are:
            2    - Flickr
            2.1  - Flickr YFCC100M dataset
            3    - Twitter
    """

    def __init__(
            self, origin_id=3, logging_level=None,
            is_local_input: bool = False, transfer_count: int = 50000,
            transferlimit: int = None,
            csv_output: bool = True, csv_suppress_linebreaks: bool = True,
            dbuser_output=None, dbserveraddress_output=None, dbname_output=None,
            dbpassword_output=None, dbserverport_output=None,
            dbuser_input=None, dbserveraddress_input=None, dbname_input=None,
            dbpassword_input=None, dbserverport_input=None):
        """Init settings for LBSNTransform"""

        # init logger level
        if logging_level is None:
            logging_level = logging.INFO
        # Set Output to Replace in case of encoding issues (console/windows)
        sys.stdout = io.TextIOWrapper(
            sys.stdout.detach(), sys.stdout.encoding, 'replace')
        sys.stdout.flush()
        self.log = HF.set_logger()

        # init global settings

        self.transfer_count = transfer_count
        self.transferlimit = None
        if transferlimit:
            self.transferlimit = transferlimit
        self.importer = HF.load_importer_mapping_module(
            origin_id)
        # get origin name and id from importer
        # e.g. yfcc100m dataset has origin id 21,
        # but is specified as general Flickr origin (2) in importer
        self.origin_id = self.importer.ORIGIN_ID
        self.origin_name = self.importer.ORIGIN_NAME
        # establish output connection
        self.dbuser_output = dbuser_output
        conn_output, cursor_output = LoadData.initialize_connection(
            dbuser_output, dbserveraddress_output,
            dbname_output, dbpassword_output, dbserverport_output)
        # store global for closing connection later
        self.cursor_output = cursor_output
        self.output = LBSNTransfer(
            db_cursor=cursor_output,
            db_connection=conn_output,
            store_csv=csv_output,
            SUPPRESS_LINEBREAKS=csv_suppress_linebreaks)
        # load from local json/csv or from PostgresDB
        self.cursor_input = None
        self.is_local_input = is_local_input
        if not self.is_local_input:
            __, cursor_input = LoadData.initialize_connection(
                dbuser_input, dbserveraddress_input,
                dbname_input, dbpassword_input, dbserverport_input,
                readonly=True)
            self.cursor_input = cursor_input
        #      loc_filelist = LoadData.read_local_files(config)
        # else:
        #      # establish input connection
        #

        # initialize stats
        self.processed_records = 0
        self.processed_total = 0
        self.skipped_low_geoaccuracy = 0
        self.initial_loop = True
        self.max_records = None
        self.how_long = None
        # field mapping structure
        # this is where all the converted data will be stored
        # note that one input record may contain many lbsn records
        self.lbsn_records = LBSNRecordDicts()

    def add_processed_records(self, lbsn_record):
        """Adds one or multiple LBSN Records (ProtoBuf)
        to collection (dicts of LBSNRecords)

        Will automatically call self.store_lbsn_records()
        """
        self.lbsn_records.add_records_to_dict(
            lbsn_record)
        self.processed_records += 1
        # On the first loop
        # or after 50.000 (default) processed records,
        # store results
        if self.initial_loop:
            self.output.store_origin(self.origin_id, self.origin_name)
            self.store_lbsn_records()
            self.initial_loop = False
        if self.processed_records >= self.transfer_count:
            self.store_lbsn_records()

    def store_lbsn_records(self):
        """Stores processed LBSN Records to chosen outpur format
        """
        self.output.store_lbsn_record_dicts(self.lbsn_records)
        self.output.commit_changes()
        self.lbsn_records.clear()
        # update statistics
        self.processed_total += self.processed_records
        self.processed_records = 0

        if self.transferlimit:
            self.max_records = self.transferlimit - self.processed_total

    def finalize_output(self):
        """finalize all transactions (csv merge etc.)
        """
        self.store_lbsn_records()
        self.output.finalize()
        # Close connections to DBs
        if not self.is_local_input:
            self.cursor_input.close()
        if self.dbuser_output:
            self.cursor_output.close()



# -*- coding: utf-8 -*-

"""
Module for mapping Twitter to common LBSN Structure.
"""

# pylint: disable=no-member

import logging
import re
import sys

import shapely.geometry as geometry
# for debugging only:
from google.protobuf import text_format
from google.protobuf.timestamp_pb2 import Timestamp
from lbsnstructure.lbsnstructure_pb2 import (CompositeKey, Language,
                                             RelationshipKey, lbsnCity,
                                             lbsnCountry, lbsnOrigin,
                                             lbsnPlace, lbsnPost,
                                             lbsnPostReaction,
                                             lbsnRelationship, lbsnUser,
                                             lbsnUserGroup)
from shapely.geometry.polygon import Polygon

from .helper_functions import HelperFunctions as HF
from .helper_functions import LBSNRecordDicts


class FieldMappingTwitter():
    """ Provides mapping function from Twitter endpoints to
        protobuf lbsnstructure
    """
    ORIGIN_NAME = "Twitter"
    ORIGIN_ID = 3

    def __init__(self,
                 disableReactionPostReferencing=False,
                 geocodes=False,
                 mapFullRelations=False,
                 map_reactions=True,
                 ignore_non_geotagged=False,
                 ignore_sources_set=set(),
                 min_geoaccuracy=None):
        # We're dealing with Twitter in this class,
        # lets create the OriginID globally
        # this OriginID is required for all CompositeKeys
        origin = lbsnOrigin()
        origin.origin_id = lbsnOrigin.TWITTER
        self.origin = origin
        # this is where all the data will be stored
        self.lbsn_records = LBSNRecordDicts()
        self.null_island = 0
        self.log = logging.getLogger('__main__')  # logging.getLogger()
        self.disable_reaction_post_referencing = disableReactionPostReferencing
        self.map_full_relations = mapFullRelations
        self.geocodes = geocodes
        self.map_reactions = map_reactions
        self.ignore_non_geotagged = ignore_non_geotagged
        self.ignore_sources_set = ignore_sources_set
        self.min_geoaccuracy = min_geoaccuracy
        self.skipped_low_geoaccuracy = 0

    def get_skipped_geoaccuracy(self):
        """Get count of records skipped due to low geoaccuracy"""
        return self.skipped_low_geoaccuracy

    def parse_json_record(self, json_string_dict, input_lbsn_type=None):
        # decide if main object is post or user json
        if input_lbsn_type and input_lbsn_type in ('friendslist',
                                                   'followerslist'):
            for user, related_user_list in json_string_dict.items():
                user_record = HF.new_lbsn_record_with_id(
                    lbsnUser(), str(user), self.origin)
                self.lbsn_records.add_records_to_dict(user_record)
                self.extract_related_users(related_user_list,
                                           input_lbsn_type, user_record)
        elif (input_lbsn_type and input_lbsn_type == 'profile') \
                or 'screen_name' in json_string_dict:
            # user
            user_record = self.extract_user(json_string_dict)
            self.lbsn_records.add_records_to_dict(user_record)
            # sys.exit(f'Post record: {text_format.MessageToString(userRecord,
            #                                                      as_utf8=True)}')
            if not user_record.is_private:
                # if user profile is private, we cannot access posts
                user_status = None
                if 'status' in json_string_dict:
                    user_status = json_string_dict.get('status')
                elif 'quoted_status' in json_string_dict:
                    user_status = json_string_dict.get('quoted_status')
                elif 'retweeted_status' in json_string_dict:
                    user_status = json_string_dict.get('retweeted_status')
                # in case user status is available
                if user_status:
                    self.parseJsonPost(user_status, userPkey=user_record.pkey)
        else:
            self.parseJsonPost(json_string_dict)

    def extract_related_users(self, related_user_list,
                              input_lbsn_type, user_record):
        for related_user in related_user_list:
            related_record = HF.new_lbsn_record_with_id(lbsnUser(),
                                                        str(related_user),
                                                        self.origin)
            self.lbsn_records.add_records_to_dict(related_record)
            # note the switch of order here,
            # direction is important for 'isConnected',
            # and the different list each give us a
            # different view on this relationship
            if input_lbsn_type == 'friendslist':
                relationship_record =\
                    HF.new_lbsn_relation_with_id(lbsnRelationship(),
                                                 user_record.pkey.id,
                                                 related_record.pkey.id,
                                                 self.origin)
            elif input_lbsn_type == 'followerslist':
                relationship_record = \
                    HF.new_lbsn_relation_with_id(lbsnRelationship(),
                                                 related_record.pkey.id,
                                                 user_record.pkey.id,
                                                 self.origin)
            relationship_record.relationship_type = \
                lbsnRelationship.isCONNECTED
            self.lbsn_records.add_relationship_to_dict(
                relationship_record)

    def parseJsonPost(self, json_string_dict, userPkey=None):
        # Post
        # 1. Extract all relevant Post Attributes
        #    1.a extract post coordinates
        #    1.b extract user attributes
        #    1.c extract place attributes
        # (poi, city, neigborhood, admin, country)
        #    1.d extract extract extended tweet,
        # if available, and extended entities, if available
        # 2. decide if post is reaction
        # (reply, quote, share, see https://developer.twitter.com/
        # en/docs/tweets/data-dictionary/overview/entities-object.html)
        # 3. if post is reaction, copy reduced reaction
        # attributes from extracted lbsnPost
        # 4. add post/reaction to recordDict
        # 5. process all referenced posts
        #    5.a Retweet(=Share) and Quote Tweets are special kinds
        # of Tweets that contain the original Tweet as an embedded object.
        #    5.b Retweets have a top-level "retweeted_status"
        # object, and Quoted Tweets have a "quoted_status" object
        # process tweet-post object
        post_record = self.extract_post(json_string_dict, userPkey)
        if post_record is None:
            # in case no post has been extracted (e.g. non_geotagged clause)
            return
        # Assignment Step
        # check if post is reaction to other post
        # reaction means: reduced structure compared to post;
        # reactions often include the complete original post,
        # therefore nested processing necessary
        if HF.is_post_reaction(json_string_dict):
            if self.map_reactions is False:
                return
            post_reaction_record = self.map_postrecord_to_postreactionrecord(
                post_record)
            refuser_pkey = None
            if 'quoted_status' in json_string_dict:
                # Note: Quote is both: Share & Reply
                if 'user' not in json_string_dict.get('quoted_status'):
                    refuser_pkey = \
                        HF.substitute_referenced_user(json_string_dict,
                                                      self.origin,
                                                      self.log)
                post_reaction_record.reaction_type = lbsnPostReaction.QUOTE
                ref_post_record = self.extract_post(
                    json_string_dict.get('quoted_status'))
            elif 'retweeted_status' in json_string_dict:
                # Note: No retweets are available when data is queried
                # using Bounding Box because of Geo-Tweet limitation:
                # "Note that native Retweets are not matched by this
                # parameter. While the original Tweet may have a location,
                # the Retweet will not"
                # see https://developer.twitter.com/en/docs/
                # tweets/filter-realtime/guides/basic-stream-parameters.html
                if 'user' not in json_string_dict.get('retweeted_status'):
                    # Current issue with Twitter search: the retweeting
                    # user is not returned in retweeted_status
                    # but we can get this from other information,
                    # such as user_mentions field from the retweet
                    # https://twittercommunity.com/t/status-retweeted-
                    # status-quoted-status-user-missing-from-search-tweets-json-response/63355
                    refuser_pkey = \
                        HF.substitute_referenced_user(json_string_dict,
                                                      self.origin,
                                                      self.log)
                post_reaction_record.reaction_type = lbsnPostReaction.SHARE
                retweet_post = json_string_dict.get('retweeted_status')
                ref_post_record = self.extract_post(retweet_post, refuser_pkey)

            elif json_string_dict.get('in_reply_to_status_id_str'):
                # if reply, original tweet is not available (?)
                post_reaction_record.reaction_type = lbsnPostReaction.COMMENT
                ref_post_record = \
                    HF.new_lbsn_record_with_id(lbsnPost(),
                                               json_string_dict.get(
                        'in_reply_to_status_id_str'),
                        self.origin)
                ref_user_record = \
                    HF.new_lbsn_record_with_id(lbsnUser(),
                                               json_string_dict.get(
                        'in_reply_to_user_id_str'),
                        self.origin)
                ref_user_record.user_name = json_string_dict.get(
                    'in_reply_to_screen_name')  # Needs to be saved
                self.lbsn_records.add_records_to_dict(ref_user_record)
                ref_post_record.user_pkey.CopyFrom(ref_user_record.pkey)

            # add referenced post pkey to reaction
            if not self.disable_reaction_post_referencing:
                post_reaction_record.referencedPost_pkey.CopyFrom(
                    ref_post_record.pkey)
                # ToDo: if a Reaction refers to another
                # reaction (Information Spread)
                # This information is currently not
                # [available from Twitter](https://developer.twitter.com/
                # en/docs/tweets/data-dictionary/overview/tweet-object):
                # "Note that retweets of retweets do not show
                # representations of the intermediary retweet [...]"
                # would be added to
                # postReactionRecord.referencedPostReaction_pkey
                if ref_post_record:
                    self.lbsn_records.add_records_to_dict(ref_post_record)
            # add postReactionRecord to Dict
            self.lbsn_records.add_records_to_dict(post_reaction_record)
        else:
            # add postReactionRecord to Dict
            self.lbsn_records.add_records_to_dict(post_record)

    def extract_user(self, json_string_dict):
        user = json_string_dict
        user_record = HF.new_lbsn_record_with_id(lbsnUser(),
                                                 user.get(
                                                 'id_str'),
                                                 self.origin)
        # get additional information about the user, if available
        user_record.user_fullname = user.get('name')
        user_record.follows = user.get('friends_count')
        user_record.is_private = user.get('protected')
        user_record.followed = user.get('followers_count')
        user_bio = user.get('description')
        if user_bio:
            user_record.biography = user_bio
        user_record.user_name = user.get('screen_name')
        listed_count = user.get('listed_count')
        if listed_count:
            user_record.group_count = listed_count
        user_record.post_count = user.get('statuses_count')
        user_record.url = f'https://twitter.com/intent/user?user_id=' \
                          f'{user_record.pkey.id}'
        ref_user_language = Language()
        ref_user_language.language_short = user.get('lang')
        user_record.user_language.CopyFrom(ref_user_language)
        user_location = user.get('location')
        if user_location:
            user_record.user_location = user_location
            if self.geocodes and user_record.user_location in self.geocodes:
                l_lat = self.geocodes[user_record.user_location][0]
                l_lng = self.geocodes[user_record.user_location][1]
                user_record.user_location_geom = "POINT(%s %s)" % (
                    l_lng, l_lat)
        # userGeoLocation = user.get('profile_location') # todo!
        user_record.liked_count = user.get('favourites_count')
        user_record.active_since.CopyFrom(
            HF.json_date_string_to_proto(user.get('created_at')))
        user_profile_image_url = user.get('profile_image_url')
        if not user_profile_image_url == f'http://abs.twimg.com/sticky/' \
                                         f'default_profile_images/' \
                                         f'default_profile_normal.png':
            user_record.profile_image_url = user_profile_image_url
        user_timezone = user.get('time_zone')
        if user_timezone:
            user_record.user_timezone = user_timezone
        user_utc_offset = user.get('utc_offset')
        if user_utc_offset:
            user_record.user_utc_offset = user_utc_offset
        # the following cannot be extracted from twitter post data
        # deutscherBundestagGroup = \
        # HF.createNewLBSNRecord_with_id(lbsnUserGroup(),
        #                               "MdB (Bundestag)",
        #                               self.origin)
        # userRecord.user_groups_member.append(
        #    deutscherBundestagGroup.pkey.id)
        # if self.mapFullRelations:
        #       relationshipRecord = \
        #       HF.createNewLBSNRelationship_with_id(lbsnRelationship(),
        #                                            userRecord.pkey.id,
        #                                            deutscherBundestagGroup.pkey.id,
        #                                            self.origin)
        #       relationshipRecord.relationship_type = lbsnRelationship.inGROUP
        #       self.lbsn_records.AddRelationshipToDict(relationshipRecord)
        # userRecord.user_groups_follows = []
        return user_record

    def extract_post(self, json_string_dict, user_pkey=None):
        post_guid = json_string_dict.get('id_str')
        if not HF.check_notice_empty_post_guid(post_guid):
            return None
        post_record = HF.new_lbsn_record_with_id(lbsnPost(),
                                                 post_guid,
                                                 self.origin)
        post_geoacc = None
        user_record = None
        user_info = json_string_dict.get('user')
        if user_info:
            # Get Post/Reaction Details of User
            user_record = self.extract_user(json_string_dict.get('user'))
        elif user_pkey:
            # userPkey is already available for posts that are statuses
            user_record = HF.new_lbsn_record_with_id(lbsnUser(),
                                                     user_pkey.id,
                                                     self.origin)
        if user_record:
            self.lbsn_records.add_records_to_dict(user_record)
        else:
            self.log.warning(f'Record {self.lbsn_records.count_glob}: '
                             f'No User record found for post: {post_guid} '
                             f'(post saved without userid)..')
            print(f'Record {self.lbsn_records.count_glob}', end='\r')
            # self.log.warning(f'{originalString}')
            # input("Press Enter to continue...
            # (post will be saved without userid)")

        # Some preprocessing for all types:
        post_coordinates = json_string_dict.get('coordinates')
        if post_coordinates:
            l_lng = post_coordinates.get('coordinates')[0]
            l_lat = post_coordinates.get('coordinates')[1]
            post_record.post_geoaccuracy = lbsnPost.LATLNG
            post_record.post_latlng = "POINT(%s %s)" % (l_lng, l_lat)

        # Check if Place is mentioned
        post_place_json = json_string_dict.get('place')
        if post_place_json:
            # we need some information from postRecord to create placeRecord
            # (e.g. user language, geoaccuracy, post_latlng)
            # some of the information from place will also modify postRecord
            # attributes; therefore return both
            if user_record:
                user_lang = user_record.user_language
            else:
                user_lang = None
            place_record, \
                post_geoacc, \
                post_country = self.extract_place(post_place_json,
                                                  post_record.post_geoaccuracy,
                                                  user_lang)
            if not post_record.post_geoaccuracy:
                post_record.post_geoaccuracy = post_geoacc
            # postRecord.post_geoaccuracy = twitterPostAttributes.geoaccuracy
            self.lbsn_records.add_records_to_dict(place_record)
            if post_country:
                post_record.country_pkey.CopyFrom(post_country.pkey)
            if isinstance(place_record, lbsnCity):
                post_record.city_pkey.CopyFrom(place_record.pkey)
            # either city or place, Twitter user cannot attach both (?)
            elif isinstance(place_record, lbsnPlace):
                post_record.place_pkey.CopyFrom(place_record.pkey)
            # substitute postRecord LatLng Coordinates from placeRecord,
            # if not already set
            if not post_record.post_latlng:
                # Note: this will also substitute Country lat/lng in post
                # this information is also available by query of
                # country_guid in posts
                # use input arg min_geoaccuracy to exclude country geo-posts
                post_record.post_latlng = place_record.geom_center
        # if still no geoinformation, send post to Null-Island
        if not post_record.post_latlng:
            if self.ignore_non_geotagged is True:
                return None
            else:
                # print(
                #     f'post_geoacc: ' \
                #     f'{lbsnPost.PostGeoaccuracy.Name(post_geoacc)}')
                # sys.exit(f'{json_string_dict}')
                self.null_island += 1
                post_record.post_latlng = "POINT(%s %s)" % (0, 0)
        if self.min_geoaccuracy:
            if not HF.geoacc_within_threshold(post_record.post_geoaccuracy,
                                              self.min_geoaccuracy):
                self.skipped_low_geoaccuracy += 1
                return
        # Process attributes of twitter post
        post_source = json_string_dict.get('source')
        if post_source:
            post_record.input_source = HF.cleanhtml(
                json_string_dict.get('source'))
            if self.ignore_sources_set and \
                    post_record.input_source in self.ignore_sources_set:
                # skip entry if in ignore list
                return None
        post_record.post_publish_date.CopyFrom(
            HF.json_date_string_to_proto(json_string_dict.get('created_at')))
        if user_record:
            post_record.user_pkey.CopyFrom(user_record.pkey)

        def value_count(x): return 0 if x is None else x
        post_record.post_quote_count = value_count(
            json_string_dict.get('quote_count'))
        post_record.post_comment_count = value_count(
            json_string_dict.get('reply_count'))
        post_record.post_share_count = value_count(
            json_string_dict.get('retweet_count'))
        post_record.post_like_count = value_count(
            json_string_dict.get('favorite_count'))
        post_record.post_url = f'https://twitter.com/statuses/{post_guid}'
        post_language = Language()
        post_language.language_short = json_string_dict.get('lang')
        post_record.post_language.CopyFrom(post_language)
        # If Extended_tweet object is available,
        # process entities and post_body (text) data from extended object
        is_truncated = json_string_dict.get('truncated')
        if is_truncated and 'extended_tweet' in json_string_dict:
            # if the "truncated" field is set to true,
            # and the "extended_tweet" object provides complete
            # "full_text" and "entities" Tweet metadata
            # Source for all data is extended object, if available
            json_string_dict = json_string_dict.get('extended_tweet')
            post_record.post_body = json_string_dict.get('full_text')
            # else:
            #    self.log.warning(f'Truncated but no extended_tweet:'
            #                     f'{json_string_dict}')
            #    input("Press Enter to continue... (entry will be skipped)")
            #    return None
        else:
            if 'full_text' in json_string_dict:
                post_record.post_body = json_string_dict.get('full_text')
            else:
                post_record.post_body = json_string_dict.get('text')
        # if 'RT @' in postRecord.post_body:
        #    sys.exit(jsonStringDict)
        if 'extended_entities' in json_string_dict:
            entities_json = json_string_dict.get('extended_entities')
        else:
            entities_json = json_string_dict.get('entities')

        hashtags_json = entities_json.get('hashtags')
        if hashtags_json:
            for hashtag in hashtags_json:  # iterate over the list
                post_record.hashtags.append(hashtag["text"])
        # Look for mentioned userRecords
        user_mentions_json = entities_json.get('user_mentions')
        if user_mentions_json:
            ref_user_records = HF.get_mentioned_users(user_mentions_json,
                                                      self.origin)
            self.lbsn_records.add_records_to_dict(ref_user_records)
            post_record.user_mentions_pkey.extend(
                [user_ref.pkey for user_ref in ref_user_records])
            if self.map_full_relations:
                self.extract_mentioned_users(
                    ref_user_records, user_record.pkey.id)
        media_json = entities_json.get('media')
        if media_json:
            post_record.post_type = HF.assign_media_post_type(media_json)
        else:
            post_record.post_type = lbsnPost.TEXT
        post_record.emoji.extend(HF.extract_emoji(post_record.post_body))
        # because standard print statement will produce escaped text,
        # we can use protobuf text_format to give us a human friendly
        # version of the text
        # log.debug(f'Post record: '
        #           f'{text_format.MessageToString(postRecord, as_utf8=True)}')
        # log.debug(f'Post record: {postRecord}')
        return post_record

    def extract_mentioned_users(self, ref_user_records, user_record_id):
        for mentioned_user_record in ref_user_records:
            relation_record = \
                HF.new_lbsn_relation_with_id(lbsnRelationship(),
                                             user_record_id,
                                             mentioned_user_record.pkey.id,
                                             self.origin)
            relation_record.relationship_type = \
                lbsnRelationship.MENTIONS_USER
            self.lbsn_records.add_relationship_to_dict(
                relation_record)

    def map_postrecord_to_postreactionrecord(self, post_record):
        post_reaction_record = lbsnPostReaction()
        post_reaction_record.pkey.CopyFrom(post_record.pkey)
        post_reaction_record.user_pkey.CopyFrom(post_record.user_pkey)
        post_reaction_record.reaction_latlng = post_record.post_latlng
        # better post_create_date, but not available from Twitter
        post_reaction_record.reaction_date.CopyFrom(
            post_record.post_publish_date)
        post_reaction_record.reaction_like_count = post_record.post_like_count
        post_reaction_record.reaction_content = post_record.post_body
        post_reaction_record.user_mentions_pkey.extend(
            [userRefPkey for userRefPkey in post_record.user_mentions_pkey])
        return post_reaction_record

    def extract_place(self, postplace_json,
                      post_geoaccuracy, user_language=None):
        place = postplace_json
        place_id = place.get('id')
        if not place_id:
            self.log.warning(f'No PlaceGuid\n\n{place}')
            input("Press Enter to continue... (entry will be skipped)")
            return None, post_geoaccuracy, None
        lon_center = 0
        lat_center = 0
        bounding_box = place.get('bounding_box')
        if bounding_box:
            bound_coordinates = bounding_box.get('coordinates')
            if bound_coordinates:
                bounding_box_points = bound_coordinates[0]
            lim_y_min, lim_y_max, lim_x_min, lim_x_max = \
                HF.get_rectangle_bounds(bounding_box_points)
            bound_points_shapely = \
                geometry.MultiPoint([(lim_x_min, lim_y_min),
                                     (lim_x_max, lim_y_max)])
            # True centroid (coords may be multipoint)
            lon_center = bound_points_shapely.centroid.coords[0][0]
            lat_center = bound_points_shapely.centroid.coords[0][1]
        place_type = place.get('place_type')
        if place_type == "country":
            # country_guid
            # in case of country,
            # we do not need to save the GUID from Twitter
            # - country_code is already unique
            country_code = place.get('country_code')
            if country_code:
                place_record = HF.new_lbsn_record_with_id(lbsnCountry(),
                                                          place.get(
                    'country_code'),
                    self.origin)
                if not post_geoaccuracy:
                    post_geoaccuracy = lbsnPost.COUNTRY
            else:
                self.log.warning(
                    f'No country_code\n\n{place}. '
                    f'PlaceEntry will be skipped..')
                return None, post_geoaccuracy, None
        elif place_type in ("city", "neighborhood", "admin"):
            # city_guid
            place_record = HF.new_lbsn_record_with_id(lbsnCity(),
                                                      place.get(
                'id'),
                self.origin)
            if not place_type == "city":
                place_record.sub_type = place_type
            if not post_geoaccuracy or post_geoaccuracy == lbsnPost.COUNTRY:
                post_geoaccuracy = lbsnPost.CITY
        elif place_type == "poi":
            # place_guid
            # For POIs, City is not available on Twitter
            place_record = HF.new_lbsn_record_with_id(lbsnPlace(),
                                                      place.get(
                'id'),
                self.origin)
            if not post_geoaccuracy or post_geoaccuracy in (lbsnPost.COUNTRY,
                                                            lbsnPost.CITY):
                post_geoaccuracy = lbsnPost.PLACE
        else:
            self.log.warning(f'No Place Type Detected: {place}')
        # for some reason, twitter place entities sometimes contain
        # linebreaks or whitespaces. We don't want this.
        place_name = place.get('name').replace('\n\r', '')
        # remove multiple whitespace
        place_name = re.sub(' +', ' ', place_name)
        if place_type == "poi" or \
           user_language is None \
           or not user_language.language_short \
           or user_language.language_short in ('en', 'und'):
            # At the moment, English name are the main references;
            # all other language specific references are stored in
            # name_alternatives - except for places, where twitter
            # has no alternative place names
            # Bugfix necessary: some English names get still saved
            # as name_alternatives
            place_record.name = place_name
        else:
            place_record.name_alternatives.append(place_name)
        place_record.url = place.get('url')
        place_record.geom_center = "POINT(%s %s)" % (lon_center, lat_center)
        if bounding_box and bound_coordinates:
            # prints: 'POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0))'
            place_record.geom_area = Polygon(bounding_box_points).wkt
        ref_country_record = None
        if not isinstance(place_record, lbsnCountry):
            ref_country_code = place.get('country_code')
            if ref_country_code:
                ref_country_record = \
                    HF.new_lbsn_record_with_id(lbsnCountry(),
                                               ref_country_code,
                                               self.origin)
                # At the moment, only English name references are processed
                if user_language is None \
                   or not user_language.language_short \
                   or user_language.language_short in ('en', 'und'):
                    ref_country_record.name = place.get(
                        'country')  # Needs to be saved
                else:
                    alt_name = place.get('country')
                    ref_country_record.name_alternatives.append(alt_name)
                self.lbsn_records.add_records_to_dict(ref_country_record)
        if post_geoaccuracy == lbsnPost.CITY and ref_country_record:
            # country_pkey only on lbsnCity(), lbsnPlace() has city_pkey,
            # but this is not available for Twitter
            place_record.country_pkey.CopyFrom(ref_country_record.pkey)
        # log.debug(f'Final Place Record: {placeRecord}')
        return place_record, post_geoaccuracy, ref_country_record



# -*- coding: utf-8 -*-

"""
Module for storing common Proto LBSN Structure to PG DB.
"""

# pylint: disable=no-member

import logging
import os
import sys
import traceback
from glob import glob
from sys import exit
import psycopg2

from lbsnstructure.lbsnstructure_pb2 import (CompositeKey, RelationshipKey,
                                             lbsnCity, lbsnCountry, lbsnPlace,
                                             lbsnPost, lbsnPostReaction,
                                             lbsnRelationship, lbsnUser,
                                             lbsnUserGroup)
from psycopg2 import sql

from .helper_functions import HelperFunctions, LBSNRecordDicts
from .shared_structure_proto_lbsndb import ProtoLBSNMapping
from .store_csv import LBSNcsv

# for debugging only:
from google.protobuf import text_format
from google.protobuf.timestamp_pb2 import Timestamp

# due to different protocol buffers implementations, import both possible types
from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
try:
    from google.protobuf.pyext._message import RepeatedCompositeContainer
except ImportError:
    pass


class LBSNTransfer():
    def __init__(self, db_cursor=None,
                 db_connection=None,
                 commit_volume=10000,
                 disable_reaction_post_ref=0,
                 store_csv=None,
                 SUPPRESS_LINEBREAKS=True):
        self.db_cursor = db_cursor
        self.db_connection = db_connection
        if not self.db_cursor:
            print("CSV Output Mode.")
        self.count_entries_commit = 0
        self.count_entries_store = 0
        self.count_affected = 0
        self.commit_volume = commit_volume
        self.store_volume = 500000
        self.count_glob = 0
        self.null_island_count = 0
        self.disable_reaction_post_ref = disable_reaction_post_ref
        self.log = logging.getLogger('__main__')
        self.batched_records = {
            lbsnCountry.DESCRIPTOR.name: list(),
            lbsnCity.DESCRIPTOR.name: list(),
            lbsnPlace.DESCRIPTOR.name: list(),
            lbsnUser.DESCRIPTOR.name: list(),
            lbsnUserGroup.DESCRIPTOR.name: list(),
            lbsnPost.DESCRIPTOR.name: list(),
            lbsnPostReaction.DESCRIPTOR.name: list(),
            lbsnRelationship.DESCRIPTOR.name: list()}

        self.count_round = 0
        # Records are batched and submitted in
        # one insert with x number of records
        self.batch_db_volume = 100
        self.store_csv = store_csv
        self.headers_written = set()
        self.db_mapping = ProtoLBSNMapping()
        # self.CSVsuppressLinebreaks = CSVsuppressLinebreaks

        if self.store_csv:
            self.csv_output = LBSNcsv(SUPPRESS_LINEBREAKS)

    def commit_changes(self):
        """Commit Changes to DB"""
        if self.db_cursor:
            self.db_connection.commit()  #
            self.count_entries_commit = 0

    def store_changes(self):
        """Write changes to CSV"""
        if self.store_csv:
            self.csv_output.clean_csv_batches(
                self.batched_records)
            self.count_entries_store = 0

    def store_origin(self, origin_id, name):
        insert_sql = \
            f'''
            INSERT INTO data."origin" (
                origin_id, name)
            VALUES ({origin_id},'{name}')
            ON CONFLICT (origin_id)
            DO NOTHING
            '''
        self.db_cursor.execute(insert_sql)

    def store_lbsn_record_dicts(self, lbsn_record_dicts):
        """Main loop for storing lbsn records to CSV or DB

        Arguments:
            field_mapping {field mapping class} -- Import Field mapping class
            with attached data

        order is important here, as PostGres will reject any
        records where Foreign Keys are violated
        therefore, records are processed starting from lowest
        granularity. Order is stored in allDicts()
        """

        self.count_round += 1
        # self.headersWritten.clear()
        r_cnt = 0
        self.count_affected = 0
        for records_dict in lbsn_record_dicts.all_dicts:
            type_name = records_dict[1]
            for record in records_dict[0].values():
                r_cnt += 1
                print(f'Storing {r_cnt} of {lbsn_record_dicts.count_glob} '
                      f'output records ({type_name})..', end='\r')
                self.prepare_lbsn_record(record, type_name)
                self.count_glob += 1  # self.dbCursor.rowcount
                self.count_entries_commit += 1  # self.dbCursor.rowcount
                self.count_entries_store += 1
                if self.db_cursor and (self.count_glob == 100 or
                                       self.count_entries_commit >
                                       self.commit_volume):
                    self.commit_changes()
                if self.store_csv and (self.count_entries_store >
                                       self.store_volume):
                    self.store_changes()
        # submit remaining rest
        self.submit_all_batches()
        # self.count_affected += x # monitoring
        print(f'\nRound {self.count_round:03d}: '
              f'Updated/Inserted {self.count_glob} records.')

    def prepare_lbsn_record(self, record, record_type):
        # clean duplicates in repeated Fields and Sort List
        self.sort_clean_proto_repeated_field(record)
        # store cleaned ProtoBuf records
        self.batched_records[record_type].append(record)
        for batch_list in self.batched_records.values():
            # if any dict contains more values than self.batchDBVolume,
            # submit/store all
            if len(batch_list) >= self.batch_db_volume:
                self.submit_all_batches()

    def submit_all_batches(self):
        for record_type, batch_list in self.batched_records.items():
            if batch_list:
                # if self.storeCSV and not record_type in self.headersWritten:
                #    self.writeCSVHeader(record_type)
                #    self.headersWritten.add(record_type)
                self.submit_lbsn_records(record_type)
                # self.funcSubmitSelector(record_type)
                batch_list.clear()

    def prepare_records(self, record_type):
        prepared_records = []
        for record in self.batched_records[record_type]:
            prepared_record = self.db_mapping.func_prepare_selector(record)
            if prepared_record:
                prepared_records.append(prepared_record)
        return prepared_records

    def submit_lbsn_records(self, record_type):
        prepared_records = self.prepare_records(record_type)
        if self.store_csv:
            self.csv_output.store_append_batch_to_csv(
                self.batched_records[record_type],
                self.count_round, record_type)
        if self.db_cursor:
            values_str = ','.join(
                [self.prepare_sqlescaped_values(record) for
                 record in prepared_records])
            insert_sql = self.insert_sql_selector(values_str, record_type)
            self.submitBatch(insert_sql)

    def insert_sql_selector(self, values_str, record_type):
        """ Select function to prepare SQL insert.

        Attributes:
            record_type     type of record
            values_str      values to be inserted
        """
        dict_switcher = {
            lbsnCountry().DESCRIPTOR.name: self.country_insertsql,
            lbsnCity().DESCRIPTOR.name: self.city_insertsql,
            lbsnPlace().DESCRIPTOR.name: self.place_insertsql,
            lbsnUser().DESCRIPTOR.name: self.user_insertsql,
            lbsnUserGroup().DESCRIPTOR.name: self.usergroup_insertsql,
            lbsnPost().DESCRIPTOR.name: self.post_insertsql,
            lbsnPostReaction().DESCRIPTOR.name: self.postreaction_insertsql,
        }
        prepare_function = dict_switcher.get(record_type)
        return prepare_function(values_str, record_type)

    def postreaction_insertsql(self, values_str, record_type):
        insert_sql = \
            f'''
            INSERT INTO data."post_reaction" (
                {self.db_mapping.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, reaction_guid)
            DO UPDATE SET
                reaction_latlng = COALESCE(
                    NULLIF(EXCLUDED.reaction_latlng,'0101000020E610000000000000000000000000000000000000'),
                    data."post_reaction".reaction_latlng,
                    '0101000020E610000000000000000000000000000000000000'),
                user_guid = COALESCE(EXCLUDED.user_guid,
                    data."post_reaction".user_guid),
                referencedPost_guid = COALESCE(EXCLUDED.referencedPost_guid,
                    data."post_reaction".referencedPost_guid),
                referencedPostreaction_guid = COALESCE(
                    EXCLUDED.referencedPostreaction_guid,
                    data."post_reaction".referencedPostreaction_guid),
                reaction_type = COALESCE(NULLIF(EXCLUDED.reaction_type, 'unknown'),
                    data."post_reaction".reaction_type, 'unknown'),
                reaction_date = COALESCE(EXCLUDED.reaction_date,
                    data."post_reaction".reaction_date),
                reaction_content = COALESCE(EXCLUDED.reaction_content,
                    data."post_reaction".reaction_content),
                reaction_like_count = COALESCE(EXCLUDED.reaction_like_count,
                    data."post_reaction".reaction_like_count),
                user_mentions = COALESCE(
                    extensions.mergeArrays(EXCLUDED.user_mentions,
                    data."post_reaction".user_mentions), ARRAY[]::text[]);
            '''
        return insert_sql

    def post_insertsql(self, values_str, record_type):
        """Insert SQL for post values

        Note COALESCE:
        - coalesce will return the first value that is not Null
        - NULLIF(value1, value2) returns null if value1 and value2 match,
            otherwise returns value1
        - combining these allows to prevent overwriting of existing
            with default values
        - if existing values are also Null, a 3rd value can be added to
            specify the final default value (e.g. the one define in
            pgtable default)
        - "default" values in postgres table are only used on insert,
            never on update (upsert)
        """
        insert_sql = \
            f'''
            INSERT INTO data."post" (
                {self.db_mapping.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, post_guid)
            DO UPDATE SET
                post_latlng = COALESCE(
                    NULLIF(EXCLUDED.post_latlng,'0101000020E610000000000000000000000000000000000000'),
                    data."post".post_latlng, '0101000020E610000000000000000000000000000000000000'),
                place_guid = COALESCE(EXCLUDED.place_guid,
                    data."post".place_guid),
                city_guid = COALESCE(EXCLUDED.city_guid,
                    data."post".city_guid),
                country_guid = COALESCE(EXCLUDED.country_guid,
                    data."post".country_guid),
                post_geoaccuracy = COALESCE(NULLIF(EXCLUDED.post_geoaccuracy,'unknown'),
                    data."post".post_geoaccuracy, 'unknown'),
                user_guid = COALESCE(EXCLUDED.user_guid,
                    data."post".user_guid),
                post_create_date = COALESCE(EXCLUDED.post_create_date,
                    data."post".post_create_date),
                post_publish_date = COALESCE(EXCLUDED.post_publish_date,
                    data."post".post_publish_date),
                post_body = COALESCE(EXCLUDED.post_body,
                    data."post".post_body),
                post_language = COALESCE(EXCLUDED.post_language,
                    data."post".post_language),
                user_mentions = COALESCE(EXCLUDED.user_mentions,
                    data."post".user_mentions),
                hashtags = COALESCE(extensions.mergeArrays(EXCLUDED.hashtags,
                    data."post".hashtags), ARRAY[]::text[]),
                emoji = COALESCE(
                    extensions.mergeArrays(EXCLUDED.emoji,
                    data."post".emoji), ARRAY[]::text[]),
                post_like_count = COALESCE(EXCLUDED.post_like_count,
                    data."post".post_like_count),
                post_comment_count = COALESCE(EXCLUDED.post_comment_count,
                    data."post".post_comment_count),
                post_views_count = COALESCE(EXCLUDED.post_views_count,
                    data."post".post_views_count),
                post_title = COALESCE(EXCLUDED.post_title,
                    data."post".post_title),
                post_thumbnail_url = COALESCE(EXCLUDED.post_thumbnail_url,
                    data."post".post_thumbnail_url),
                post_url = COALESCE(EXCLUDED.post_url,
                    data."post".post_url),
                post_type = COALESCE(NULLIF(EXCLUDED.post_type, 'text'),
                    data."post".post_type, 'text'),
                post_filter = COALESCE(EXCLUDED.post_filter,
                    data."post".post_filter),
                post_quote_count = COALESCE(EXCLUDED.post_quote_count,
                    data."post".post_quote_count),
                post_share_count = COALESCE(EXCLUDED.post_share_count,
                    data."post".post_share_count),
                input_source = COALESCE(EXCLUDED.input_source,
                    data."post".input_source),
                post_content_license = COALESCE(
                    EXCLUDED.post_content_license,
                        data."post".post_content_license);
            '''
        return insert_sql

    def user_insertsql(self, values_str, record_type):
        insert_sql = \
            f'''
            INSERT INTO data."user" (
                {self.db_mapping.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, user_guid)
            DO UPDATE SET
                user_name = COALESCE(EXCLUDED.user_name,
                    data."user".user_name),
                user_fullname = COALESCE(EXCLUDED.user_fullname,
                    data."user".user_fullname),
                follows = GREATEST(COALESCE(
                    EXCLUDED.follows, data."user".follows),
                    COALESCE(data."user".follows, EXCLUDED.follows)),
                followed = GREATEST(COALESCE(
                    EXCLUDED.followed, data."user".followed),
                    COALESCE(data."user".followed, EXCLUDED.followed)),
                group_count = GREATEST(COALESCE(
                    EXCLUDED.group_count, data."user".group_count),
                    COALESCE(data."user".group_count, EXCLUDED.group_count)),
                biography = COALESCE(EXCLUDED.biography,
                    data."user".biography),
                post_count = GREATEST(COALESCE(
                    EXCLUDED.post_count, "user".post_count),
                    COALESCE(data."user".post_count, EXCLUDED.post_count)),
                is_private = COALESCE(EXCLUDED.is_private,
                    data."user".is_private),
                url = COALESCE(EXCLUDED.url, data."user".url),
                is_available = COALESCE(EXCLUDED.is_available,
                    data."user".is_available),
                user_language = COALESCE(EXCLUDED.user_language,
                    data."user".user_language),
                user_location = COALESCE(EXCLUDED.user_location,
                    data."user".user_location),
                user_location_geom = COALESCE(EXCLUDED.user_location_geom,
                    data."user".user_location_geom),
                liked_count = GREATEST(COALESCE(
                    EXCLUDED.liked_count, data."user".liked_count),
                    COALESCE(data."user".liked_count, EXCLUDED.liked_count)),
                active_since = COALESCE(EXCLUDED.active_since,
                    data."user".active_since),
                profile_image_url = COALESCE(EXCLUDED.profile_image_url,
                    data."user".profile_image_url),
                user_timezone = COALESCE(EXCLUDED.user_timezone,
                    data."user".user_timezone),
                user_utc_offset = COALESCE(EXCLUDED.user_utc_offset,
                    data."user".user_utc_offset),
                user_groups_member = COALESCE(
                    extensions.mergeArrays(EXCLUDED.user_groups_member,
                    data."user".user_groups_member), ARRAY[]::text[]),
                user_groups_follows = COALESCE(
                    extensions.mergeArrays(EXCLUDED.user_groups_follows,
                    data."user".user_groups_follows), ARRAY[]::text[]);
            '''
        return insert_sql

    def usergroup_insertsql(self, values_str, record_type):
        insert_sql = \
            f'''
            INSERT INTO data."user_groups" (
                {self.db_mapping.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, usergroup_guid)
            DO UPDATE SET
                usergroup_name = COALESCE(EXCLUDED.usergroup_name,
                    data."user_groups".usergroup_name),
                usergroup_description = COALESCE(
                    EXCLUDED.usergroup_description,
                    data."user_groups".usergroup_description),
                member_count = GREATEST(COALESCE(
                    EXCLUDED.member_count, data."user_groups".member_count),
                    COALESCE(data."user_groups".member_count,
                    EXCLUDED.member_count)),
                usergroup_createdate = COALESCE(
                    EXCLUDED.usergroup_createdate,
                    data."user_groups".usergroup_createdate),
                user_owner = COALESCE(EXCLUDED.user_owner,
                    data."user_groups".user_owner);
            '''
        # No coalesce for user: in case user changes or
        # removes information, this should also be removed from the record
        return insert_sql

    def place_insertsql(self, values_str, record_type):
        insert_sql = \
            f'''
            INSERT INTO data."place" (
                {self.db_mapping.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id,place_guid)
            DO UPDATE SET
                name = COALESCE(EXCLUDED.name, data."place".name),
                name_alternatives = COALESCE((
                    SELECT array_remove(altNamesNewArray,data."place".name)
                    from extensions.mergeArrays(EXCLUDED.name_alternatives,
                    data."place".name_alternatives) AS altNamesNewArray),
                    ARRAY[]::text[]),
                geom_center = COALESCE(EXCLUDED.geom_center,
                    data."place".geom_center),
                geom_area = COALESCE(EXCLUDED.geom_area,
                    data."place".geom_area),
                url = COALESCE(EXCLUDED.url, data."place".url),
                city_guid = COALESCE(EXCLUDED.city_guid,
                    data."place".city_guid),
                post_count = GREATEST(COALESCE(EXCLUDED.post_count,
                    data."place".post_count), COALESCE(
                        data."place".post_count, EXCLUDED.post_count));
            '''
        return insert_sql

    def city_insertsql(self, values_str, record_type):
        insert_sql = \
            f'''
            INSERT INTO data."city" (
                {self.db_mapping.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id,city_guid)
            DO UPDATE SET
                name = COALESCE(EXCLUDED.name, data."city".name),
                name_alternatives = COALESCE((
                    SELECT array_remove(altNamesNewArray,data."city".name)
                    from extensions.mergeArrays(EXCLUDED.name_alternatives,
                    data."city".name_alternatives) AS altNamesNewArray),
                    ARRAY[]::text[]),
                geom_center = COALESCE(EXCLUDED.geom_center,
                    data."city".geom_center),
                geom_area = COALESCE(EXCLUDED.geom_area,
                    data."city".geom_area),
                url = COALESCE(EXCLUDED.url, data."city".url),
                country_guid = COALESCE(EXCLUDED.country_guid,
                    data."city".country_guid),
                sub_type = COALESCE(EXCLUDED.sub_type, data."city".sub_type);
            '''
        return insert_sql

    def country_insertsql(self, values_str, record_type):
        insert_sql = \
            f'''
            INSERT INTO data."country" (
                {self.db_mapping.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id,country_guid)
            DO UPDATE SET
                name = COALESCE(EXCLUDED.name, data."country".name),
                name_alternatives = COALESCE((
                    SELECT array_remove(altNamesNewArray,data."country".name)
                    from extensions.mergeArrays(EXCLUDED.name_alternatives,
                    data."country".name_alternatives) AS altNamesNewArray),
                    ARRAY[]::text[]),
                geom_center = COALESCE(EXCLUDED.geom_center,
                    data."country".geom_center),
                geom_area = COALESCE(EXCLUDED.geom_area,
                    data."country".geom_area),
                url = COALESCE(EXCLUDED.url, data."country".url);
            '''
        # Array merge of alternatives:
        # Arrays cannot be null, therefore COALESCE(
        # [if array not null],[otherwise create empty array])
        # We don't want the english name to appear in alternatives,
        # therefore: array_remove(altNamesNewArray,"country".name)
        # Finally, merge New Entries with existing ones (distinct):
        # extensions.mergeArrays([new],[old]) uses custom mergeArrays
        # function (see function definitions)
        return insert_sql

    def submitLbsnRelationships(self):
        """submit relationships of different types

        record[1] is the PostgresQL formatted list of values,
        record[0] is the type of relationship that determines
            the table selection
        """
        selectFriends = [relationship[1] for relationship in
                         self.batched_records[lbsnRelationship(
                         ).DESCRIPTOR.name] if relationship[0] == "isfriend"]
        if selectFriends:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    selectFriends, self.count_round, '_user_friends_user')
            if self.db_cursor:
                args_isFriend = ','.join(selectFriends)
                insert_sql = \
                    f'''
                    INSERT INTO relations."_user_friends_user" (
                        {self.typeNamesHeaderDict["_user_friends_user"]})
                    VALUES {args_isFriend}
                    ON CONFLICT (origin_id, user_guid, friend_guid)
                    DO NOTHING
                    '''
                self.submitBatch(insert_sql)
        selectConnected = [relationship[1] for relationship in
                           self.batched_records[lbsnRelationship(
                           ).DESCRIPTOR.name] if
                           relationship[0] == "isconnected"]
        if selectConnected:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    selectConnected, self.count_round, '_user_connectsto_user')
            if self.db_cursor:
                args_isConnected = ','.join(selectConnected)
                insert_sql = \
                    f'''
                        INSERT INTO relations."_user_connectsto_user" (
                            {self.typeNamesHeaderDict["_user_connectsto_user"]})
                        VALUES {args_isConnected}
                        ON CONFLICT (origin_id, user_guid,
                            connectedto_user_guid)
                        DO NOTHING
                    '''
                self.submitBatch(insert_sql)
        selectUserGroupMember = [relationship[1] for relationship in
                                 self.batched_records[lbsnRelationship(
                                 ).DESCRIPTOR.name] if
                                 relationship[0] == "ingroup"]
        if selectUserGroupMember:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    selectUserGroupMember, self.count_round,
                    '_user_memberof_group')
            if self.db_cursor:
                args_isInGroup = ','.join(selectUserGroupMember)
                insert_sql = \
                    f'''
                    INSERT INTO relations."_user_memberof_group" (
                        {self.typeNamesHeaderDict["_user_memberof_group"]})
                    VALUES {args_isInGroup}
                    ON CONFLICT (origin_id, user_guid, group_guid)
                    DO NOTHING
                    '''
                self.submitBatch(insert_sql)
        selectUserGroupMember = [relationship[1] for relationship in
                                 self.batched_records[lbsnRelationship(
                                 ).DESCRIPTOR.name] if
                                 relationship[0] == "followsgroup"]
        if selectUserGroupMember:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    selectUserGroupMember, self.count_round,
                    '_user_follows_group')
            if self.db_cursor:
                args_isInGroup = ','.join(selectUserGroupMember)
                insert_sql = \
                    f'''
                    INSERT INTO relations."_user_follows_group" (
                        {self.typeNamesHeaderDict["_user_follows_group"]})
                    VALUES {args_isInGroup}
                    ON CONFLICT (origin_id, user_guid, group_guid)
                    DO NOTHING
                    '''
                self.submitBatch(insert_sql)
        selectUserMentions = [relationship[1] for relationship in
                              self.batched_records[lbsnRelationship(
                              ).DESCRIPTOR.name] if
                              relationship[0] == "mentions_user"]
        if selectUserMentions:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    selectUserMentions, self.count_round,
                    '_user_mentions_user')
            if self.db_cursor:
                args_isInGroup = ','.join(selectUserMentions)
                insert_sql = \
                    f'''
                    INSERT INTO relations."_user_mentions_user" (
                        {self.typeNamesHeaderDict["_user_mentions_user"]})
                    VALUES {args_isInGroup}
                    ON CONFLICT (origin_id, user_guid, mentioneduser_guid)
                    DO NOTHING
                    '''
                self.submitBatch(insert_sql)

    def submitBatch(self, insert_sql):
        """Submit Batch to PG DB.

        Needs testing: is using Savepoint for each insert slower
        than rolling back entire commit?
        for performance, see https://stackoverflow.com/questions/
            12206600/how-to-speed-up-insertion-performance-in-postgresql
        or this: https://stackoverflow.com/questions/8134602/
        psycopg2-insert-multiple-rows-with-one-query
        """
        self.db_cursor.execute("SAVEPOINT submit_recordBatch")
        tsuccessful = False
        while not tsuccessful:
            try:
                self.db_cursor.execute(insert_sql)
            except psycopg2.IntegrityError as e:
                if '(post_language)' in e.diag.message_detail or \
                        '(user_language)' in e.diag.message_detail:
                    # If language does not exist, we'll trust Twitter
                    # and add this to our language list
                    missingLanguage = e.diag.message_detail.partition(
                        "language)=(")[2].partition(") is not present")[0]
                    print(
                        f'TransactionIntegrityError, inserting language "'
                        f'{missingLanguage}" first..               ')
                    self.db_cursor.execute(
                        "ROLLBACK TO SAVEPOINT submit_recordBatch")
                    insert_language_sql = '''
                           INSERT INTO data."language"
                            (language_short,language_name,language_name_de)
                           VALUES (%s,NULL,NULL);
                           '''
                    self.db_cursor.execute(
                        insert_language_sql, (missingLanguage,))
                else:
                    sys.exit(f'{e}')
            except psycopg2.DataError as e:
                sys.exit(f'{e}\nINSERT SQL WAS: {insert_sql}')
            except ValueError as e:
                self.log.warning(f'{e}')
                input("Press Enter to continue... (entry will be skipped)")
                self.log.warning(f'{insert_sql}')
                input("args:... ")
                # self.log.warning(f'{args_str}')
                self.db_cursor.execute(
                    "ROLLBACK TO SAVEPOINT submit_recordBatch")
                tsuccessful = True
            except psycopg2.ProgrammingError as e:
                sys.exit(f'{e}\nINSERT SQL WAS: {insert_sql}')
            else:
                # self.count_affected += self.dbCursor.rowcount # monitoring
                self.db_cursor.execute("RELEASE SAVEPOINT submit_recordBatch")
                tsuccessful = True

    def prepare_sqlescaped_values(self, *args):
        # dynamically construct sql value injection
        # e.g. record_sql = '''(%s,%s,%s,%s,%s,%s,%s)'''
        record_sql = f'''{','.join('%s' for x in range(0, len(args)))}'''
        # inject values
        preparedSQLRecord = self.db_cursor.mogrify(record_sql, tuple(args))
        # mogrify returns a byte object,
        # we decode it so it can be used as a string again
        preparedSQLRecord = preparedSQLRecord.decode()
        return preparedSQLRecord

    def sort_clean_proto_repeated_field(self, record):
        """Remove duplicate values in repeated field, sort alphabetically

        ProtocolBuffers has no unique list field type. This function will
        remove duplicates, which is needed for unique compare.

        There is a 'bug' in Python implementation of ProtocolBuffers:
        - depending on the implementation type in use, it is possible
        to spot either 'RepeatedCompositeFieldContainer'
            or 'RepeatedCompositeContainer'
        - solution here: import and compare to both types
        - this is not ideal, since both types are internal to PB and
            subject to change
        - see [proto-bug](https://github.com/protocolbuffers/
            protobuf/issues/3870)
        """
        for descriptor in record.DESCRIPTOR.fields:
            if descriptor.label == descriptor.LABEL_REPEATED:
                x = getattr(record, descriptor.name)
                if x and not (isinstance(x, RepeatedCompositeFieldContainer) or
                              isinstance(x, RepeatedCompositeContainer)):
                    xCleaned = set(x)
                    xSorted = sorted(xCleaned)
                    # Complete clear of repeated field
                    for _ in range(0, len(x)):
                        x.pop()
                    # add sorted list
                    x.extend(xSorted)

    def finalize(self):
        """ Final procedure calls:
            - clean and merge csv batches
        """
        if self.store_csv:
            self.csv_output.clean_csv_batches(self.batched_records)



# -*- coding: utf-8 -*-

"""
Module for mapping Flickr to common LBSN Structure.
"""

# pylint: disable=no-member

import logging
from decimal import Decimal

# for debugging only:
from google.protobuf import text_format
from lbsnstructure.lbsnstructure_pb2 import (CompositeKey, Language,
                                             RelationshipKey, lbsnCity,
                                             lbsnCountry, lbsnOrigin,
                                             lbsnPlace, lbsnPost,
                                             lbsnPostReaction,
                                             lbsnRelationship, lbsnUser,
                                             lbsnUserGroup)

from .helper_functions import HelperFunctions as HF
from .helper_functions import LBSNRecordDicts


class FieldMappingFlickr():
    """ Provides mapping function from Flickr endpoints to
        protobuf lbsnstructure
    """
    ORIGIN_NAME = "Flickr"
    ORIGIN_ID = 2

    def __init__(self,
                 disable_reaction_post_referencing=False,
                 geocodes=False,
                 map_full_relations=False,
                 map_reactions=True,
                 ignore_non_geotagged=False,
                 ignore_sources_set=set(),
                 min_geoaccuracy=None):
        # We're dealing with Flickr in this class, lets create the OriginID
        # globally
        # this OriginID is required for all CompositeKeys
        origin = lbsnOrigin()
        origin.origin_id = lbsnOrigin.FLICKR
        self.origin = origin
        self.null_island = 0
        # this is where all the data will be stored
        self.lbsn_records = LBSNRecordDicts()
        self.log = logging.getLogger('__main__')  # get the main logger object
        self.skipped_count = 0
        self.skipped_low_geoaccuracy = 0
        # self.disableReactionPostReferencing = disableReactionPostReferencing
        # self.mapFullRelations = mapFullRelations
        # self.geocodes = geocodes

    def parse_csv_record(self, record):
        """Entry point for flickr CSV data:
            - Decide if CSV record contains user-info or post-info
            - Skip empty or malformed records

        Attributes:
        record    A single row from CSV, stored as list type.
        """
        if len(record) < 12 or not record[0].isdigit():
            # skip
            self.skipped_count += 1
            return
        else:
            self.extract_flickr_post(record)

    def extract_flickr_post(self, record):
        """Main function for processing Flickr CSV entry.
           This mothod is adapted to a special structure, adapt if needed.

        To Do:
            - parameterize column numbers and structure
            - provide external config-file for specific CSV structures
            - currently not included in lbsn mapping are MachineTags,
              GeoContext (indoors, outdoors), WoeId
              and some extra attributes only present for Flickr
        """
        post_guid = record[5]
        if not HF.check_notice_empty_post_guid(post_guid):
            return None
        post_record = HF.new_lbsn_record_with_id(lbsnPost(),
                                                 post_guid,
                                                 self.origin)
        user_record = HF.new_lbsn_record_with_id(lbsnUser(),
                                                 record[7],
                                                 self.origin)
        user_record.user_name = record[6]
        user_record.url = f'http://www.flickr.com/photos/{user_record.pkey.id}/'
        if user_record:
            post_record.user_pkey.CopyFrom(user_record.pkey)
        self.lbsn_records.add_records_to_dict(user_record)
        post_record.post_latlng = self.flickr_extract_postlatlng(record)
        geoaccuracy = FieldMappingFlickr.flickr_map_geoaccuracy(record[13])
        if geoaccuracy:
            post_record.post_geoaccuracy = geoaccuracy
        if record[19]:
            # we need some information from postRecord to create placeRecord
            # (e.g.  user language, geoaccuracy, post_latlng)
            # some of the information from place will also modify postRecord
            place_record = HF.new_lbsn_record_with_id(lbsnPlace(),
                                                      record[19],
                                                      self.origin)
            self.lbsn_records.add_records_to_dict(place_record)
            post_record.place_pkey.CopyFrom(place_record.pkey)
        post_record.post_publish_date.CopyFrom(
            HF.parse_csv_datestring_to_protobuf(record[9]))
        post_record.post_create_date.CopyFrom(
            HF.parse_csv_datestring_to_protobuf(record[8]))
        # valueCount = lambda x: 0 if x is None else x

        def value_count(x): return int(x) if x.isdigit() else 0
        post_record.post_views_count = value_count(record[10])
        post_record.post_comment_count = value_count(record[18])
        post_record.post_like_count = value_count(record[17])
        post_record.post_url = f'http://flickr.com/photo.gne?id={post_guid}'
        post_record.post_body = FieldMappingFlickr.reverse_csv_comma_replace(
            record[21])
        post_record.post_title = FieldMappingFlickr.reverse_csv_comma_replace(
            record[3])
        post_record.post_thumbnail_url = record[4]
        record_tags_list = list(filter(None, record[11].split(";")))
        if record_tags_list:
            for tag in record_tags_list:
                tag = FieldMappingFlickr.clean_tags_from_flickr(tag)
                post_record.hashtags.append(tag)
        record_media_type = record[16]
        if record_media_type and record_media_type == "video":
            post_record.post_type = lbsnPost.VIDEO
        else:
            post_record.post_type = lbsnPost.IMAGE
        post_record.post_content_license = value_count(record[14])
        self.lbsn_records.add_records_to_dict(post_record)

    @staticmethod
    def reverse_csv_comma_replace(csv_string):
        """Flickr CSV data is stored without quotes by replacing commas with semicolon.
        This functon will reverse replacement.
        """
        return csv_string.replace(";", ",")

    @staticmethod
    def clean_tags_from_flickr(tag):
        """Clean special vars not allowed in tags.
        """
        characters_to_replace = ('{', '}')
        for char_check in characters_to_replace:
            tag = tag.replace(char_check, '')
        return tag

    @staticmethod
    def flickr_map_geoaccuracy(flickr_geo_accuracy_level):
        """Flickr Geoaccuracy Levels (16) are mapped to four LBSNstructure levels:
           LBSN PostGeoaccuracy: UNKNOWN = 0; LATLNG = 1; PLACE = 2; CITY = 3;
           COUNTRY = 4
           Fickr: World level is 1, Country is ~3, Region ~6, City ~11,
           Street ~16.

           Flickr Current range is 1-16. Defaults to 16 if not specified.

        Attributes:
        flickr_geo_accuracy_level   Geoaccuracy Level returned from Flickr
        (String, e.g.: "Level12")
        """
        lbsn_geoaccuracy = False
        stripped_level = flickr_geo_accuracy_level.lstrip('Level').strip()
        if stripped_level.isdigit():
            stripped_level = int(stripped_level)
            if stripped_level >= 15:
                lbsn_geoaccuracy = lbsnPost.LATLNG
            elif stripped_level >= 12:
                lbsn_geoaccuracy = lbsnPost.PLACE
            elif stripped_level >= 8:
                lbsn_geoaccuracy = lbsnPost.CITY
            else:
                lbsn_geoaccuracy = lbsnPost.COUNTRY
        else:
            if flickr_geo_accuracy_level == "Street":
                lbsn_geoaccuracy = lbsnPost.LATLNG
            elif flickr_geo_accuracy_level in ("City", "Region"):
                lbsn_geoaccuracy = lbsnPost.CITY
            elif flickr_geo_accuracy_level in ("Country", "World"):
                lbsn_geoaccuracy = lbsnPost.COUNTRY
        return lbsn_geoaccuracy

    def flickr_extract_postlatlng(self, record):
        """Basic routine for extracting lat/lng coordinates from post.
           - checks for consistency and errors
           - in case of any issue, entry is submitted to Null island (0, 0)
        """
        lat_entry = record[1]
        lng_entry = record[2]
        if lat_entry == "" and lng_entry == "":
            l_lat, l_lng = 0, 0
        else:
            try:
                l_lng = Decimal(lng_entry)
                l_lat = Decimal(lat_entry)
            except:
                l_lat, l_lng = 0, 0

        if (l_lat == 0 and l_lng == 0) \
                or l_lat > 90 or l_lat < -90 \
                or l_lng > 180 or l_lng < -180:
            l_lat, l_lng = 0, 0
            self.send_to_null_island(lat_entry, lng_entry, record[5])
        return FieldMappingFlickr.lat_lng_to_wkt(l_lat, l_lng)

    @staticmethod
    def lat_lng_to_wkt(lat, lng):
        """Convert lat lng to WKT (Well-Known-Text)
        """
        point_latlng_string = "POINT(%s %s)" % (lng, lat)
        return point_latlng_string

    def send_to_null_island(self, lat_entry, lng_entry, record_guid):
        """Logs entries with problematic lat/lng's,
           increases Null Island Counter by 1.
        """
        self.log.debug(f'NULL island: Guid {record_guid} - '
                       f'Coordinates: {lat_entry}, {lng_entry}')
        self.null_island += 1
